@@ -1,86 +1,48 @@
 from __future__ import annotations
 
 from dataclasses import dataclass
-from typing import Union
+from typing import Annotated, Union
 
 from xdsl.dialects.builtin import (ContainerOf, Float16Type, Float64Type, IndexType,
-                                   IntegerType, Float32Type, IntegerAttr)
-from xdsl.ir import MLContext, Operation, SSAValue, Attribute
-from xdsl.irdl import (AnyOf, irdl_op_definition, AttributeDef, AnyAttr,
-                       ResultDef, OperandDef)
+                                   IntegerType, Float32Type, IntegerAttr, FloatAttr,
+                                   Attribute, AnyFloat)
+from xdsl.ir import Operation, SSAValue, Dialect, OpResult
+from xdsl.irdl import (AnyOf, irdl_op_definition, OpAttr, AnyAttr,
+                       Operand)
 from xdsl.utils.exceptions import VerifyException
 
 signlessIntegerLike = ContainerOf(AnyOf([IntegerType, IndexType]))
 floatingPointLike = ContainerOf(AnyOf([Float16Type, Float32Type, Float64Type]))
 
 
-@dataclass
-class Arith:
-    ctx: MLContext
-
-    def __post_init__(self):
-        self.ctx.register_op(Constant)
-
-        # Integer-like
-        self.ctx.register_op(Addi)
-        self.ctx.register_op(Subi)
-        self.ctx.register_op(Muli)
-        self.ctx.register_op(DivUI)
-        self.ctx.register_op(DivSI)
-        self.ctx.register_op(FloorDivSI)
-        self.ctx.register_op(CeilDivSI)
-        self.ctx.register_op(CeilDivUI)
-        self.ctx.register_op(RemUI)
-        self.ctx.register_op(RemSI)
-        self.ctx.register_op(MinSI)
-        self.ctx.register_op(MaxSI)
-        self.ctx.register_op(MinUI)
-        self.ctx.register_op(MaxUI)
-
-        # Float-like
-        self.ctx.register_op(Addf)
-        self.ctx.register_op(Subf)
-        self.ctx.register_op(Mulf)
-        self.ctx.register_op(Divf)
-
-        # Comparison/Condition
-        self.ctx.register_op(Cmpi)
-        self.ctx.register_op(Select)
-
-        # Logical
-        self.ctx.register_op(AndI)
-        self.ctx.register_op(OrI)
-        self.ctx.register_op(XOrI)
-
-        # Shift
-        self.ctx.register_op(ShLI)
-        self.ctx.register_op(ShRUI)
-        self.ctx.register_op(ShRSI)
-
-        # Min/Max
-        self.ctx.register_op(Minf)
-        self.ctx.register_op(Maxf)
-
 
 @irdl_op_definition
 class Constant(Operation):
     name: str = "arith.constant"
-    result = ResultDef(AnyAttr())
-    value = AttributeDef(AnyAttr())
+    result: Annotated[OpResult, AnyAttr()]
+    value: OpAttr[Attribute]
 
     @staticmethod
     def from_attr(attr: Attribute, typ: Attribute) -> Constant:
         return Constant.create(result_types=[typ], attributes={"value": attr})
 
     @staticmethod
-    def from_int_constant(val: Union[int, Attribute],
-                          typ: Union[int, Attribute]) -> Constant:
+    def from_int_and_width(val: Union[int, Attribute],
+                           typ: Union[int, Attribute]) -> Constant:
         if isinstance(typ, int):
             typ = IntegerType.from_width(typ)
         return Constant.create(
             result_types=[typ],
             attributes={"value": IntegerAttr.from_params(val, typ)})
 
+    # To add tests for this constructor
+    @staticmethod
+    def from_float_and_width(val: Union[float, Attribute],
+                             typ: AnyFloat) -> Constant:
+        return Constant.create(
+            result_types=[typ],
+            attributes={"value": FloatAttr.from_value(val, typ)})
+
 
 @dataclass
 class BinaryOperation(Operation):
@@ -99,9 +61,9 @@ class BinaryOperation(Operation):
 @irdl_op_definition
 class Addi(BinaryOperation):
     name: str = "arith.addi"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -114,9 +76,9 @@ class Addi(BinaryOperation):
 @irdl_op_definition
 class Muli(BinaryOperation):
     name: str = "arith.muli"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -129,9 +91,9 @@ class Muli(BinaryOperation):
 @irdl_op_definition
 class Subi(BinaryOperation):
     name: str = "arith.subi"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -142,16 +104,16 @@ class Subi(BinaryOperation):
 
 
 @irdl_op_definition
-class DivUI(Operation):
+class DivUI(BinaryOperation):
     """
     Unsigned integer division. Rounds towards zero. Treats the leading bit as
     the most significant, i.e. for `i16` given two's complement representation,
     `6 / -2 = 6 / (2^16 - 2) = 0`.
     """
     name: str = "arith.divui"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -168,23 +130,27 @@ class DivSI(BinaryOperation):
     sign, i.e. `6 / -2 = -3`.
     """
     name: str = "arith.divsi"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
             operand2: Union[Operation, SSAValue]) -> DivSI:
+        operand1 = SSAValue.get(operand1)
         return DivSI.build(operands=[operand1, operand2],
                            result_types=[operand1.typ])
 
 
 @irdl_op_definition
 class FloorDivSI(BinaryOperation):
+    """
+    Signed floor integer division. Rounds towards negative infinity i.e. `5 / -2 = -3`.
+    """
     name: str = "arith.floordivsi"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -197,9 +163,9 @@ class FloorDivSI(BinaryOperation):
 @irdl_op_definition
 class CeilDivSI(BinaryOperation):
     name: str = "arith.ceildivsi"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -212,9 +178,9 @@ class CeilDivSI(BinaryOperation):
 @irdl_op_definition
 class CeilDivUI(BinaryOperation):
     name: str = "arith.ceildivui"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -227,9 +193,9 @@ class CeilDivUI(BinaryOperation):
 @irdl_op_definition
 class RemUI(BinaryOperation):
     name: str = "arith.remui"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -242,9 +208,9 @@ class RemUI(BinaryOperation):
 @irdl_op_definition
 class RemSI(BinaryOperation):
     name: str = "arith.remsi"
-    lhs = OperandDef(IntegerType)
-    rhs = OperandDef(IntegerType)
-    result = ResultDef(IntegerType)
+    lhs: Annotated[Operand, IntegerType]
+    rhs: Annotated[Operand, IntegerType]
+    result: Annotated[OpResult, IntegerType]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -257,9 +223,9 @@ class RemSI(BinaryOperation):
 @irdl_op_definition
 class MinUI(BinaryOperation):
     name: str = "arith.minui"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -272,9 +238,9 @@ class MinUI(BinaryOperation):
 @irdl_op_definition
 class MaxUI(BinaryOperation):
     name: str = "arith.maxui"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -287,9 +253,9 @@ class MaxUI(BinaryOperation):
 @irdl_op_definition
 class MinSI(BinaryOperation):
     name: str = "arith.minsi"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -302,9 +268,9 @@ class MinSI(BinaryOperation):
 @irdl_op_definition
 class MaxSI(BinaryOperation):
     name: str = "arith.maxsi"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -317,9 +283,9 @@ class MaxSI(BinaryOperation):
 @irdl_op_definition
 class AndI(BinaryOperation):
     name: str = "arith.andi"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -332,9 +298,9 @@ class AndI(BinaryOperation):
 @irdl_op_definition
 class OrI(BinaryOperation):
     name: str = "arith.ori"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -347,9 +313,9 @@ class OrI(BinaryOperation):
 @irdl_op_definition
 class XOrI(BinaryOperation):
     name: str = "arith.xori"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -366,9 +332,9 @@ class ShLI(Operation):
     amount. The low order bits are filled with zeros.
     """
     name: str = "arith.shli"
-    lhs = OperandDef(IntegerType)
-    rhs = OperandDef(IntegerType)
-    result = ResultDef(IntegerType)
+    lhs: Annotated[Operand, IntegerType]
+    rhs: Annotated[Operand, IntegerType]
+    result: Annotated[OpResult, IntegerType]
 
     # TODO replace with trait
     def verify_(self) -> None:
@@ -392,9 +358,9 @@ class ShRUI(Operation):
     always filled with zeros.
     """
     name: str = "arith.shrui"
-    lhs = OperandDef(signlessIntegerLike)
-    rhs = OperandDef(signlessIntegerLike)
-    result = ResultDef(signlessIntegerLike)
+    lhs: Annotated[Operand, signlessIntegerLike]
+    rhs: Annotated[Operand, signlessIntegerLike]
+    result: Annotated[OpResult, signlessIntegerLike]
 
     # TODO replace with trait
     def verify_(self) -> None:
@@ -419,9 +385,9 @@ class ShRSI(Operation):
     value (which means that the sign of the value is preserved).
     """
     name: str = "arith.shrsi"
-    lhs = OperandDef(IntegerType)
-    rhs = OperandDef(IntegerType)
-    result = ResultDef(IntegerType)
+    lhs: Annotated[Operand, IntegerType]
+    rhs: Annotated[Operand, IntegerType]
+    result: Annotated[OpResult, IntegerType]
 
     # TODO replace with trait
     def verify_(self) -> None:
@@ -457,10 +423,10 @@ class Cmpi(Operation):
     -   unsigned greater than or equal (mnemonic: `"uge"`; integer value: `9`)
     """
     name: str = "arith.cmpi"
-    predicate = AttributeDef(IntegerAttr)
-    lhs = OperandDef(IntegerType)
-    rhs = OperandDef(IntegerType)
-    result = ResultDef(IntegerType.from_width(1))
+    predicate: OpAttr[IntegerAttr]
+    lhs: Annotated[Operand, IntegerType]
+    rhs: Annotated[Operand, IntegerType]
+    result: Annotated[OpResult, IntegerType.from_width(1)]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -508,10 +474,10 @@ class Select(Operation):
     The second and the third operand must have the same type.
     """
     name: str = "arith.select"
-    cond = OperandDef(IntegerType.from_width(1))  # should be unsigned
-    lhs = OperandDef(Attribute)
-    rhs = OperandDef(Attribute)
-    result = ResultDef(Attribute)
+    cond: Annotated[Operand, IntegerType.from_width(1)]  # should be unsigned
+    lhs: Annotated[Operand, Attribute]
+    rhs: Annotated[Operand, Attribute]
+    result: Annotated[OpResult, Attribute]
 
     # TODO replace with trait
     def verify_(self) -> None:
@@ -533,9 +499,9 @@ class Select(Operation):
 @irdl_op_definition
 class Addf(BinaryOperation):
     name: str = "arith.addf"
-    lhs = OperandDef(floatingPointLike)
-    rhs = OperandDef(floatingPointLike)
-    result = ResultDef(floatingPointLike)
+    lhs: Annotated[Operand, floatingPointLike]
+    rhs: Annotated[Operand, floatingPointLike]
+    result: Annotated[OpResult, floatingPointLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -548,9 +514,9 @@ class Addf(BinaryOperation):
 @irdl_op_definition
 class Subf(BinaryOperation):
     name: str = "arith.subf"
-    lhs = OperandDef(floatingPointLike)
-    rhs = OperandDef(floatingPointLike)
-    result = ResultDef(floatingPointLike)
+    lhs: Annotated[Operand, floatingPointLike]
+    rhs: Annotated[Operand, floatingPointLike]
+    result: Annotated[OpResult, floatingPointLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -563,9 +529,9 @@ class Subf(BinaryOperation):
 @irdl_op_definition
 class Mulf(BinaryOperation):
     name: str = "arith.mulf"
-    lhs = OperandDef(floatingPointLike)
-    rhs = OperandDef(floatingPointLike)
-    result = ResultDef(floatingPointLike)
+    lhs: Annotated[Operand, floatingPointLike]
+    rhs: Annotated[Operand, floatingPointLike]
+    result: Annotated[OpResult, floatingPointLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -578,9 +544,9 @@ class Mulf(BinaryOperation):
 @irdl_op_definition
 class Divf(BinaryOperation):
     name: str = "arith.divf"
-    lhs = OperandDef(floatingPointLike)
-    rhs = OperandDef(floatingPointLike)
-    result = ResultDef(floatingPointLike)
+    lhs: Annotated[Operand, floatingPointLike]
+    rhs: Annotated[Operand, floatingPointLike]
+    result: Annotated[OpResult, floatingPointLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -593,9 +559,9 @@ class Divf(BinaryOperation):
 @irdl_op_definition
 class Maxf(BinaryOperation):
     name: str = "arith.maxf"
-    lhs = OperandDef(floatingPointLike)
-    rhs = OperandDef(floatingPointLike)
-    result = ResultDef(floatingPointLike)
+    lhs: Annotated[Operand, floatingPointLike]
+    rhs: Annotated[Operand, floatingPointLike]
+    result: Annotated[OpResult, floatingPointLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -608,9 +574,9 @@ class Maxf(BinaryOperation):
 @irdl_op_definition
 class Minf(BinaryOperation):
     name: str = "arith.minf"
-    lhs = OperandDef(floatingPointLike)
-    rhs = OperandDef(floatingPointLike)
-    result = ResultDef(floatingPointLike)
+    lhs: Annotated[Operand, floatingPointLike]
+    rhs: Annotated[Operand, floatingPointLike]
+    result: Annotated[OpResult, floatingPointLike]
 
     @staticmethod
     def get(operand1: Union[Operation, SSAValue],
@@ -618,3 +584,47 @@ class Minf(BinaryOperation):
         operand1 = SSAValue.get(operand1)
         return Minf.build(operands=[operand1, operand2],
                           result_types=[operand1.typ])
+
+Arith = Dialect([
+        Constant,
+
+        # Integer-like
+        Addi,
+        Subi,
+        Muli,
+        DivUI,
+        DivSI,
+        FloorDivSI,
+        CeilDivSI,
+        CeilDivUI,
+        RemUI,
+        RemSI,
+        MinSI,
+        MaxSI,
+        MinUI,
+        MaxUI,
+
+        # Float-like
+        Addf,
+        Subf,
+        Mulf,
+        Divf,
+
+        # Comparison/Condition
+        Cmpi,
+        Select,
+
+        # Logical
+        AndI,
+        OrI,
+        XOrI,
+
+        # Shift
+        ShLI,
+        ShRUI,
+        ShRSI,
+
+        # Min/Max
+        Minf,
+        Maxf],
+        [])
