@@ -2,64 +2,31 @@ from __future__ import annotations
 
 from dataclasses import dataclass
 from enum import Enum
-from typing import (TypeAlias, List, cast, Type, Sequence, Optional,
-                    TYPE_CHECKING, Any, TypeVar)
+from typing import (TypeAlias, List, cast, Type, Sequence, TYPE_CHECKING, Any,
+                    TypeVar)
 
-from xdsl.ir import (MLContext, Data, MLIRType, ParametrizedAttribute,
-                     Operation, SSAValue, Region, Attribute)
-from xdsl.irdl import (AttributeDef, VarOperandDef, VarRegionDef, VarResultDef,
+from xdsl.ir import (Data, MLIRType, ParametrizedAttribute, Operation, Region,
+                     Attribute, Dialect)
+from xdsl.irdl import (OpAttr, VarOpResult, VarOperand, VarRegion,
                        irdl_attr_definition, attr_constr_coercion,
                        irdl_data_definition, irdl_to_attr_constraint,
                        irdl_op_definition, builder, ParameterDef,
-                       SingleBlockRegionDef, Generic, GenericData,
-                       AttrConstraint, AnyAttr)
+                       SingleBlockRegion, Generic, GenericData, AttrConstraint,
+                       AnyAttr)
 from xdsl.utils.exceptions import VerifyException
 
 if TYPE_CHECKING:
-    from xdsl.parser import Parser, ParserError
+    from xdsl.parser import BaseParser
+    from utils.exceptions import ParseError
     from xdsl.printer import Printer
 
 
-@dataclass
-class Builtin:
-    ctx: MLContext
-
-    def __post_init__(self):
-        self.ctx.register_attr(StringAttr)
-        self.ctx.register_attr(FlatSymbolRefAttr)
-        self.ctx.register_attr(SymbolNameAttr)
-        self.ctx.register_attr(IntAttr)
-        self.ctx.register_attr(IntegerAttr)
-        self.ctx.register_attr(ArrayAttr)
-        self.ctx.register_attr(VectorType)
-        self.ctx.register_attr(TensorType)
-        self.ctx.register_attr(UnrankedTensorType)
-        self.ctx.register_attr(DenseIntOrFPElementsAttr)
-        self.ctx.register_attr(UnitAttr)
-        self.ctx.register_attr(TupleType)
-
-        self.ctx.register_attr(FunctionType)
-        self.ctx.register_attr(Float16Type)
-        self.ctx.register_attr(Float32Type)
-        self.ctx.register_attr(Float64Type)
-        self.ctx.register_attr(FloatData)
-        self.ctx.register_attr(FloatAttr)
-        self.ctx.register_attr(IntegerType)
-        self.ctx.register_attr(IndexType)
-
-        self.ctx.register_attr(NoneAttr)
-        self.ctx.register_attr(OpaqueAttr)
-
-        self.ctx.register_op(ModuleOp)
-        self.ctx.register_op(UnregisteredOp)
-
-
 @irdl_attr_definition
 class StringAttr(Data[str]):
     name = "string"
 
     @staticmethod
-    def parse_parameter(parser: Parser) -> str:
+    def parse_parameter(parser: BaseParser) -> str:
         data = parser.parse_str_literal()
         return data
 
@@ -115,7 +82,7 @@ class IntAttr(Data[int]):
     name = "int"
 
     @staticmethod
-    def parse_parameter(parser: Parser) -> int:
+    def parse_parameter(parser: BaseParser) -> int:
         data = parser.parse_int_literal()
         return data
 
@@ -144,14 +111,14 @@ class SignednessAttr(Data[Signedness]):
     name = "signedness"
 
     @staticmethod
-    def parse_parameter(parser: Parser) -> Signedness:
+    def parse_parameter(parser: BaseParser) -> Signedness:
         if parser.parse_optional_string("signless") is not None:
             return Signedness.SIGNLESS
         elif parser.parse_optional_string("signed") is not None:
             return Signedness.SIGNED
         elif parser.parse_optional_string("unsigned") is not None:
             return Signedness.UNSIGNED
-        raise ParserError(parser.get_pos(), "Expected signedness")
+        raise ParseError(parser.get_pos(), "Expected signedness")
 
     @staticmethod
     def print_parameter(data: Signedness, printer: Printer) -> None:
@@ -260,7 +227,7 @@ class FloatData(Data[float]):
     name = "float_data"
 
     @staticmethod
-    def parse_parameter(parser: Parser) -> float:
+    def parse_parameter(parser: BaseParser) -> float:
         return parser.parse_float_literal()
 
     @staticmethod
@@ -333,7 +300,7 @@ class ArrayAttr(GenericData[List[_ArrayAttrT]]):
     name = "array"
 
     @staticmethod
-    def parse_parameter(parser: Parser) -> List[_ArrayAttrT]:
+    def parse_parameter(parser: BaseParser) -> List[_ArrayAttrT]:
         parser.parse_char("[")
         data = parser.parse_list(parser.parse_optional_attribute)
         parser.parse_char("]")
@@ -377,6 +344,60 @@ class ArrayAttr(GenericData[List[_ArrayAttrT]]):
 AnyArrayAttr: TypeAlias = ArrayAttr[Attribute]
 
 
+@irdl_attr_definition
+class DictionaryAttr(GenericData[dict[str, Attribute]]):
+    name = "dictionary"
+
+    @staticmethod
+    def parse_parameter(parser: BaseParser) -> dict[str, Attribute]:
+        # force MLIR style parsing of attribute
+        from xdsl.parser import MLIRParser
+        return MLIRParser.parse_optional_attr_dict(parser)
+
+    @staticmethod
+    def print_parameter(data: dict[str, Attribute], printer: Printer) -> None:
+        printer.print_string("{")
+        printer.print_dictionary(data, printer.print_string_literal,
+                                 printer.print_attribute)
+        printer.print_string("}")
+
+    @staticmethod
+    def generic_constraint_coercion(args: tuple[Any]) -> AttrConstraint:
+        raise Exception(f"Unsupported operation on {DictionaryAttr.name}")
+
+    def verify(self) -> None:
+        if not isinstance(self.data, dict):
+            raise VerifyException(
+                f"Wrong type given to attribute {self.name}: got"
+                f" {type(self.data)}, but expected dictionary of"
+                " attributes")
+        for key, val in self.data.items():
+            if not isinstance(key, str):
+                raise VerifyException(
+                    f"{self.name} key expects str, but {key} "
+                    f"element is of type {type(key)}")
+            if not isinstance(val, Attribute):
+                raise VerifyException(
+                    f"{self.name} key expects attribute, but {val} "
+                    f"element is of type {type(val)}")
+
+    @staticmethod
+    @builder
+    def from_dict(data: dict[str | StringAttr, Attribute]) -> DictionaryAttr:
+        to_add_data: dict[str, Attribute] = {}
+        for k, v in data.items():
+            # try to coerce keys into StringAttr
+            if isinstance(k, StringAttr):
+                k = k.data
+            # if coercion fails, raise KeyError!
+            if not isinstance(k, str):
+                raise TypeError(
+                    f"DictionaryAttr.from_dict expects keys to"
+                    f" be of type str or StringAttr, but {type(k)} provided")
+            to_add_data[k] = v
+        return DictionaryAttr(to_add_data)
+
+
 @irdl_attr_definition
 class TupleType(ParametrizedAttribute):
     name = "tuple"
@@ -407,12 +428,10 @@ class VectorType(Generic[_VectorTypeElems], ParametrizedAttribute, MLIRType):
 
     @staticmethod
     @builder
-    def from_type_and_list(
+    def from_element_type_and_shape(
         referenced_type: _VectorTypeElems,
-        shape: Optional[List[int | IntegerAttr[IndexType]]] = None
+        shape: List[int | IntegerAttr[IndexType]]
     ) -> VectorType[_VectorTypeElems]:
-        if shape is None:
-            shape = [1]
         return VectorType([
             ArrayAttr.from_list(
                 [IntegerAttr[IntegerType].build(d) for d in shape]),
@@ -451,7 +470,7 @@ class TensorType(Generic[_TensorTypeElems], ParametrizedAttribute, MLIRType):
     @builder
     def from_type_and_list(
         referenced_type: _TensorTypeElems,
-        shape: Optional[Sequence[int | IntegerAttr[IndexType]]] = None
+        shape: Sequence[int | IntegerAttr[IndexType]] | None = None
     ) -> TensorType[_TensorTypeElems]:
         if shape is None:
             shape = [1]
@@ -551,7 +570,7 @@ class DenseIntOrFPElementsAttr(ParametrizedAttribute):
 
     @staticmethod
     @builder
-    def from_index_list(
+    def create_dense_index(
             type: VectorOrTensorOf[IndexType],
             data: List[int | IntegerAttr[IndexType]]
     ) -> DenseIntOrFPElementsAttr:
@@ -563,7 +582,7 @@ class DenseIntOrFPElementsAttr(ParametrizedAttribute):
 
     @staticmethod
     @builder
-    def from_int_list(
+    def create_dense_int(
         type: VectorOrTensorOf[IntegerType],
         data: List[int | IntegerAttr[IntegerType]]
     ) -> DenseIntOrFPElementsAttr:
@@ -575,28 +594,28 @@ class DenseIntOrFPElementsAttr(ParametrizedAttribute):
 
     @staticmethod
     @builder
-    def from_float_list(
+    def create_dense_float(
             type: VectorOrTensorOf[AnyFloat],
-            data: List[float | AnyFloatAttr]) -> DenseIntOrFPElementsAttr:
+            data: List[int | float | AnyFloatAttr]
+    ) -> DenseIntOrFPElementsAttr:
         data_attr = [
-            FloatAttr.from_value(d, type.element_type)
-            if isinstance(d, float) else d for d in data
+            FloatAttr.from_value(float(d), type.element_type)
+            if not isinstance(d, FloatAttr) else d for d in data
         ]
         return DenseIntOrFPElementsAttr([type, ArrayAttr.from_list(data_attr)])
 
     @staticmethod
     @builder
     def from_list(
-        type: VectorOrTensorOf[Attribute],
-        data: List[int | IntegerAttr[IntegerType]]
-        | List[int | IntegerAttr[IndexType]] | List[float | AnyFloatAttr]
+        type: VectorOrTensorOf[Attribute], data: List[int | AnyIntegerAttr]
+        | List[int | float | AnyFloatAttr]
     ) -> DenseIntOrFPElementsAttr:
         if isinstance(type.element_type, IntegerType):
-            return DenseIntOrFPElementsAttr.from_int_list(type, data)
+            return DenseIntOrFPElementsAttr.create_dense_int(type, data)
         elif isinstance(type.element_type, IndexType):
-            return DenseIntOrFPElementsAttr.from_index_list(type, data)
+            return DenseIntOrFPElementsAttr.create_dense_index(type, data)
         elif isinstance(type.element_type, AnyFloat):
-            return DenseIntOrFPElementsAttr.from_float_list(type, data)
+            return DenseIntOrFPElementsAttr.create_dense_float(type, data)
         else:
             raise TypeError(f"Unsupported element type {type.element_type}")
 
@@ -606,7 +625,7 @@ class DenseIntOrFPElementsAttr(ParametrizedAttribute):
             data: List[int] | List[float],
             typ: IntegerType | IndexType | AnyFloat
     ) -> DenseIntOrFPElementsAttr:
-        t = AnyVectorType.from_type_and_list(typ, [len(data)])
+        t = AnyVectorType.from_element_type_and_shape(typ, [len(data)])
         return DenseIntOrFPElementsAttr.from_list(t, data)
 
     @staticmethod
@@ -629,7 +648,7 @@ class FunctionType(ParametrizedAttribute, MLIRType):
     @staticmethod
     @builder
     def from_lists(inputs: List[Attribute],
-                   outputs: List[Attribute]) -> Attribute:
+                   outputs: List[Attribute]) -> FunctionType:
         return FunctionType(
             [ArrayAttr.from_list(inputs),
              ArrayAttr.from_list(outputs)])
@@ -637,7 +656,7 @@ class FunctionType(ParametrizedAttribute, MLIRType):
     @staticmethod
     @builder
     def from_attrs(inputs: ArrayAttr[Attribute],
-                   outputs: ArrayAttr[Attribute]) -> Attribute:
+                   outputs: ArrayAttr[Attribute]) -> FunctionType:
         return FunctionType([inputs, outputs])
 
 
@@ -667,36 +686,37 @@ class OpaqueAttr(ParametrizedAttribute):
 class UnregisteredOp(Operation):
     name: str = "builtin.unregistered"
 
-    op_name__ = AttributeDef(StringAttr)
-    args = VarOperandDef(AnyAttr())
-    res = VarResultDef(AnyAttr())
-    regs = VarRegionDef()
+    op_name__: OpAttr[StringAttr]
+    args: VarOperand
+    res: VarOpResult
+    regs: VarRegion
 
     @property
     def op_name(self) -> StringAttr:
         return self.op_name__  # type: ignore
 
-    @staticmethod
-    def from_name(name: str | StringAttr,
-                  args: list[SSAValue | Operation] = [],
-                  res: list[Attribute] = [],
-                  regs: list[Region] = [],
-                  attrs: dict[str, Attribute] = {}) -> UnregisteredOp:
-        if "op_name__" in attrs:
-            raise Exception(
-                "Cannot create an unregistered op with an __op_name attribute")
-        attrs["op_name__"] = StringAttr.build(name)
-        return UnregisteredOp.build(operands=args,
-                                    result_types=res,
-                                    regions=regs,
-                                    attributes=attrs)
+    @classmethod
+    def with_name(cls, name: str, ctx: MLContext) -> type[Operation]:
+        if name in ctx.registered_unregistered_ops:
+            return ctx.registered_unregistered_ops[name]  # type: ignore
+
+        class UnregisteredOpWithName(UnregisteredOp):
+
+            @classmethod
+            def create(cls, **kwargs):
+                op = super().create(**kwargs)
+                op.attributes['op_name__'] = StringAttr.build(name)
+                return op
+
+        ctx.registered_unregistered_ops[name] = UnregisteredOpWithName
+        return UnregisteredOpWithName
 
 
 @irdl_op_definition
 class ModuleOp(Operation):
     name: str = "builtin.module"
 
-    body = SingleBlockRegionDef()
+    body: SingleBlockRegion
 
     @property
     def ops(self) -> List[Operation]:
@@ -720,3 +740,33 @@ class ModuleOp(Operation):
 f16 = Float16Type()
 f32 = Float32Type()
 f64 = Float64Type()
+
+Builtin = Dialect(
+    [ModuleOp, UnregisteredOp],
+    [
+        StringAttr,
+        FlatSymbolRefAttr,
+        SymbolNameAttr,
+        IntAttr,
+        IntegerAttr,
+        ArrayAttr,
+        DictionaryAttr,
+        DenseIntOrFPElementsAttr,
+        UnitAttr,
+        FloatData,
+        NoneAttr,
+        OpaqueAttr,
+
+        # Types
+        FunctionType,
+        Float16Type,
+        Float32Type,
+        Float64Type,
+        FloatAttr,
+        TupleType,
+        IntegerType,
+        IndexType,
+        VectorType,
+        TensorType,
+        UnrankedTensorType
+    ])
