@@ -1,25 +1,65 @@
 from __future__ import annotations
 
+import re
 from abc import ABC, abstractmethod
 from dataclasses import dataclass, field
 from frozenlist import FrozenList
-from typing import (TYPE_CHECKING, Any, Callable, Generic, Protocol, Sequence,
-                    TypeVar, cast)
+from io import StringIO
+from typing import (TYPE_CHECKING, Any, Callable, Generic, Optional, Protocol,
+                    Sequence, TypeVar, cast, Iterator, Union, ClassVar)
+import sys
 
 # Used for cyclic dependencies in type hints
 if TYPE_CHECKING:
-    from xdsl.parser import Parser
+    from xdsl.parser import Parser, BaseParser
     from xdsl.printer import Printer
     from xdsl.irdl import OpDef, ParamAttrDef
 
 OpT = TypeVar('OpT', bound='Operation')
 
 
+@dataclass
+class Dialect:
+    """Contains the operations and attributes of a specific dialect"""
+    _operations: list[type[Operation]] = field(default_factory=list,
+                                               init=True,
+                                               repr=True)
+    _attributes: list[type[Attribute]] = field(default_factory=list,
+                                               init=True,
+                                               repr=True)
+
+    @property
+    def operations(self) -> Iterator[type[Operation]]:
+        return iter(self._operations)
+
+    @property
+    def attributes(self) -> Iterator[type[Attribute]]:
+        return iter(self._attributes)
+
+    def __call__(self, ctx: MLContext) -> None:
+        print(
+            "Calling a dialect in order to register it is deprecated and will soon be removed.",
+            file=sys.stderr)
+        # TODO; Remove this function in a future release.
+        assert isinstance(ctx, MLContext)
+        ctx.register_dialect(self)
+
+
 @dataclass
 class MLContext:
     """Contains structures for operations/attributes registration."""
     _registeredOps: dict[str, type[Operation]] = field(default_factory=dict)
     _registeredAttrs: dict[str, type[Attribute]] = field(default_factory=dict)
+    registered_unregistered_ops: dict[str, type[Operation]] = field(
+        default_factory=dict)
+
+    def register_dialect(self, dialect: Dialect):
+        """Register a dialect. Operation and Attribute names should be unique"""
+        for op in dialect.operations:
+            self.register_op(op)
+
+        for attr in dialect.attributes:
+            self.register_attr(attr)
 
     def register_op(self, op: type[Operation]) -> None:
         """Register an operation definition. Operation names should be unique."""
@@ -81,7 +121,19 @@ class SSAValue(ABC):
     uses: set[Use] = field(init=False, default_factory=set, repr=False)
     """All uses of the value."""
 
-    name: str | None = field(init=False, default=None)
+    _name: str | None = field(init=False, default=None)
+
+    _name_regex: ClassVar[re.Pattern] = re.compile(
+        r'[A-Za-z0-9._$-]*[A-Za-z._$-]')
+
+    @property
+    def name(self) -> str | None:
+        return self._name
+
+    @name.setter
+    def name(self, name: str):
+        if self._name_regex.fullmatch(name):
+            self._name = name
 
     @staticmethod
     def get(arg: SSAValue | Operation) -> SSAValue:
@@ -229,6 +281,13 @@ class Attribute(ABC):
         """
         pass
 
+    def __str__(self) -> str:
+        from xdsl.printer import Printer
+        res = StringIO()
+        printer = Printer(stream=res)
+        printer.print_attribute(self)
+        return res.getvalue()
+
 
 DataElement = TypeVar("DataElement")
 
@@ -240,7 +299,7 @@ class Data(Generic[DataElement], Attribute, ABC):
 
     @staticmethod
     @abstractmethod
-    def parse_parameter(parser: Parser) -> DataElement:
+    def parse_parameter(parser: BaseParser) -> DataElement:
         """Parse the attribute parameter."""
 
     @staticmethod
@@ -255,7 +314,7 @@ class ParametrizedAttribute(Attribute):
     parameters: list[Attribute] = field(default_factory=list)
 
     @staticmethod
-    def parse_parameters(parser: Parser) -> list[Attribute]:
+    def parse_parameters(parser: BaseParser) -> list[Attribute]:
         """Parse the attribute parameters."""
         return parser.parse_paramattr_parameters()
 
@@ -271,21 +330,30 @@ class ParametrizedAttribute(Attribute):
 
 
 @dataclass
-class IRNode(object):
+class IRNode(ABC):
 
-    def is_ancestor(cls: IRNode, op: IRNode) -> bool:
+    parent: IRNode | None
+
+    def is_ancestor(self, op: IRNode) -> bool:
         "Returns true if the IRNode is an ancestor of another IRNode."
-        if op is cls:
+        if op is self:
             return True
         if op.parent is None:
             return False
-        return cls.is_ancestor(op.parent)
+        return self.is_ancestor(op.parent)
 
-    def get_toplevel_object(cls: IRNode) -> IRNode:
+    def get_toplevel_object(self) -> IRNode:
         """Get the operation, block, or region ancestor that has no parents."""
-        if cls.parent is None:
-            return cls
-        return cls.parent.get_toplevel_object()
+        if self.parent is None:
+            return self
+        return self.parent.get_toplevel_object()
+
+    def is_structurally_equivalent(
+            self,
+            other: IRNode,
+            context: Optional[dict[IRNode, IRNode]] = None) -> bool:
+        """Check if two IR nodes are structurally equivalent."""
+        ...
 
 
 @dataclass
@@ -317,16 +385,14 @@ class Operation(IRNode):
     """The block containing this operation."""
 
     def parent_op(self) -> Operation | None:
-        try:
-            return self.parent_region().parent
-        except AttributeError:
-            None
+        if p := self.parent_region():
+            return p.parent
+        return None
 
     def parent_region(self) -> Region | None:
-        try:
-            return self.parent_block().parent
-        except AttributeError:
-            None
+        if p := self.parent_block():
+            return p.parent
+        return None
 
     def parent_block(self) -> Block | None:
         return self.parent
@@ -446,7 +512,7 @@ class Operation(IRNode):
 
     @classmethod
     def parse(cls: type[_OperationType], result_types: list[Attribute],
-              parser: Parser) -> _OperationType:
+              parser: BaseParser) -> _OperationType:
         return parser.parse_op_with_default_format(cls, result_types)
 
     def print(self, printer: Printer):
@@ -513,6 +579,48 @@ class Operation(IRNode):
             raise Exception("Cannot detach a toplevel operation.")
         self.parent.detach_op(self)
 
+    def is_structurally_equivalent(
+            self,
+            other: IRNode,
+            context: Optional[dict[IRNode, IRNode]] = None) -> bool:
+        """
+        Check if two operations are structurally equivalent.
+        The context is a mapping of IR nodes to IR nodes that are already known to be equivalent.
+        This enables checking whether the use dependencies and successors are equivalent.
+        """
+        if context is None:
+            context = {}
+        if not isinstance(other, Operation):
+            return False
+        if self.name != other.name:
+            return False
+        if len(self.operands) != len(other.operands) or \
+           len(self.results) != len(other.results) or \
+           len(self.regions) != len(other.regions) or \
+           len(self.successors) != len(other.successors) or \
+            self.attributes != other.attributes:
+            return False
+        if self.parent and other.parent and context.get(
+                self.parent) != other.parent:
+            return False
+        if not all(
+                context.get(operand) == other_operand for operand,
+                other_operand in zip(self.operands, other.operands)):
+            return False
+        if not all(
+                context.get(successor) == other_successor for successor,
+                other_successor in zip(self.successors, other.successors)):
+            return False
+        if not all(
+                region.is_structurally_equivalent(other_region, context)
+                for region, other_region in zip(self.regions, other.regions)):
+            return False
+        # Add results of this operation to the context
+        for result, other_result in zip(self.results, other.results):
+            context[result] = other_result
+
+        return True
+
     def __eq__(self, other: object) -> bool:
         return self is other
 
@@ -530,6 +638,8 @@ class Operation(IRNode):
 class Block(IRNode):
     """A sequence of operations"""
 
+    declared_at: 'Span' | None = None
+
     _args: FrozenList[BlockArgument] = field(default_factory=FrozenList,
                                              init=False)
     """The basic block arguments."""
@@ -737,6 +847,35 @@ class Block(IRNode):
         for op in self.ops:
             op.erase(safe_erase=safe_erase, drop_references=False)
 
+    def is_structurally_equivalent(
+            self,
+            other: IRNode,
+            context: Optional[dict[IRNode, IRNode]] = None) -> bool:
+        """
+        Check if two blocks are structurally equivalent.
+        The context is a mapping of IR nodes to IR nodes that are already known to be equivalent.
+        This enables checking whether the use dependencies and successors are equivalent.
+        """
+        if context is None:
+            context = {}
+        if not isinstance(other, Block):
+            return False
+        if len(self.args) != len(other.args) or \
+            len(self.ops) != len(other.ops):
+            return False
+        for arg, other_arg in zip(self.args, other.args):
+            if arg.typ != other_arg.typ:
+                return False
+            context[arg] = other_arg
+        # Add self to the context so Operations can check for identical parents
+        context[self] = other
+        if not all(
+                op.is_structurally_equivalent(other_op, context)
+                for op, other_op in zip(self.ops, other.ops)):
+            return False
+
+        return True
+
     def __eq__(self, other: object) -> bool:
         return self is other
 
@@ -865,8 +1004,6 @@ class Region(IRNode):
         else:
             block_idx = block
             block = self.blocks[block_idx]
-        if block.parent is not self:
-            raise Exception("Cannot detach block from a different region.")
         block.parent = None
         self.blocks = self.blocks[:block_idx] + self.blocks[block_idx + 1:]
         return block
@@ -944,3 +1081,28 @@ class Region(IRNode):
         self.blocks = []
         for block in region.blocks:
             block.parent = region
+
+    def is_structurally_equivalent(
+            self,
+            other: IRNode,
+            context: Optional[dict[IRNode, IRNode]] = None) -> bool:
+        """
+        Check if two regions are structurally equivalent.
+        The context is a mapping of IR nodes to IR nodes that are already known to be equivalent.
+        This enables checking whether the use dependencies and successors are equivalent.
+        """
+        if context is None:
+            context = {}
+        if not isinstance(other, Region):
+            return False
+        if len(self.blocks) != len(other.blocks):
+            return False
+        # register all blocks in the context so we can check whether ops have
+        # the corrects successors
+        for block, other_block in zip(self.blocks, other.blocks):
+            context[block] = other_block
+        if not all(
+                block.is_structurally_equivalent(other_block, context)
+                for block, other_block in zip(self.blocks, other.blocks)):
+            return False
+        return True
