@@ -5,6 +5,7 @@ from abc import ABC, abstractmethod
 from dataclasses import dataclass, field
 from enum import Enum
 from frozenlist import FrozenList
+from functools import reduce
 from inspect import isclass
 from typing import (Annotated, Any, Callable, Generic, Sequence, TypeAlias,
                     TypeVar, Union, cast, get_args, get_origin, get_type_hints)
@@ -13,7 +14,7 @@ from types import UnionType, GenericAlias
 from xdsl.ir import (Attribute, Block, Data, OpResult, Operation,
                      ParametrizedAttribute, Region, SSAValue)
 from xdsl.utils.diagnostic import Diagnostic
-from xdsl.utils.exceptions import VerifyException
+from xdsl.utils.exceptions import BuilderNotFoundException, VerifyException
 from xdsl.utils.hints import is_satisfying_hint
 
 # pyright: reportMissingParameterType=false, reportUnknownParameterType=false
@@ -27,6 +28,9 @@ def error(op: Operation, msg: str):
 
 class IRDLAnnotations(Enum):
     ParamDefAnnot = 1
+    AttributeDefAnnot = 2
+    OptAttributeDefAnnot = 3
+    SingleBlockRegionAnnot = 4
 
 
 #   ____                _             _       _
@@ -95,8 +99,7 @@ class AnyAttr(AttrConstraint):
     """Constraint that is verified by all attributes."""
 
     def verify(self, attr: Attribute) -> None:
-        if not isinstance(attr, Attribute):
-            raise VerifyException(f"Expected attribute, but got {attr}")
+        pass
 
 
 @dataclass(init=False)
@@ -114,8 +117,6 @@ class AnyOf(AttrConstraint):
 
     def verify(self, attr: Attribute) -> None:
         for attr_constr in self.attr_constrs:
-            if isinstance(attr_constr, Attribute) and attr_constr == attr:
-                return
             try:
                 attr_constr.verify(attr)
                 return
@@ -143,13 +144,13 @@ class ParamAttrConstraint(AttrConstraint):
     and also constrain its parameters with additional constraints.
     """
 
-    base_attr: type[Attribute]
+    base_attr: type[ParametrizedAttribute]
     """The base attribute type."""
 
     param_constrs: list[AttrConstraint]
     """The attribute parameter constraints"""
 
-    def __init__(self, base_attr: type[Attribute],
+    def __init__(self, base_attr: type[ParametrizedAttribute],
                  param_constrs: Sequence[(Attribute | type[Attribute]
                                           | AttrConstraint)]):
         self.base_attr = base_attr
@@ -158,13 +159,9 @@ class ParamAttrConstraint(AttrConstraint):
         ]
 
     def verify(self, attr: Attribute) -> None:
-        assert isinstance(attr, ParametrizedAttribute)
         if not isinstance(attr, self.base_attr):
-            # the type checker concludes that attr has type 'Never', therefore the cast
-            name = cast(Attribute, attr).name
             raise VerifyException(
-                f"Base attribute {self.base_attr.name} expected, but got {name}"
-            )
+                f"{attr} should be of base attribute {self.base_attr.name}")
         if len(self.param_constrs) != len(attr.parameters):
             raise VerifyException(
                 f"{len(self.param_constrs)} parameters expected, "
@@ -182,6 +179,9 @@ def irdl_to_attr_constraint(
     if isinstance(irdl, AttrConstraint):
         return irdl
 
+    if isinstance(irdl, Attribute):
+        return EqAttrConstraint(irdl)
+
     # Annotated case
     # Each argument of the Annotated type corresponds to a constraint to satisfy.
     if get_origin(irdl) == Annotated:
@@ -360,16 +360,25 @@ class OperandDef(OperandOrResultDef):
         self.constr = attr_constr_coercion(typ)
 
 
+Operand: TypeAlias = SSAValue
+
+
 @dataclass(init=False)
 class VarOperandDef(OperandDef, VariadicDef):
     """An IRDL variadic operand definition."""
 
 
+VarOperand: TypeAlias = list[SSAValue]
+
+
 @dataclass(init=False)
 class OptOperandDef(VarOperandDef, OptionalDef):
     """An IRDL optional operand definition."""
 
 
+OptOperand: TypeAlias = SSAValue | None
+
+
 @dataclass(init=False)
 class ResultDef(OperandOrResultDef):
     """An IRDL result definition."""
@@ -386,11 +395,17 @@ class VarResultDef(ResultDef, VariadicDef):
     """An IRDL variadic result definition."""
 
 
+VarOpResult: TypeAlias = list[OpResult]
+
+
 @dataclass(init=False)
 class OptResultDef(VarResultDef, OptionalDef):
     """An IRDL optional result definition."""
 
 
+OptOpResult: TypeAlias = OpResult | None
+
+
 @dataclass
 class RegionDef(Region):
     """
@@ -408,6 +423,10 @@ class OptRegionDef(RegionDef, OptionalDef):
     """An IRDL optional region definition."""
 
 
+VarRegion: TypeAlias = list[Region]
+OptRegion: TypeAlias = Region | None
+
+
 @dataclass
 class SingleBlockRegionDef(RegionDef):
     """An IRDL region definition that expects exactly one block."""
@@ -421,6 +440,14 @@ class OptSingleBlockRegionDef(RegionDef, OptionalDef):
     """An IRDL optional region definition that expects exactly one block."""
 
 
+SingleBlockRegion: TypeAlias = Annotated[
+    Region, IRDLAnnotations.SingleBlockRegionAnnot]
+VarSingleBlockRegion: TypeAlias = Annotated[
+    list[Region], IRDLAnnotations.SingleBlockRegionAnnot]
+OptSingleBlockRegion: TypeAlias = Annotated[
+    Region | None, IRDLAnnotations.SingleBlockRegionAnnot]
+
+
 @dataclass(init=False)
 class AttributeDef:
     """An IRDL attribute definition."""
@@ -440,6 +467,13 @@ class OptAttributeDef(AttributeDef):
         super().__init__(typ)
 
 
+_OpAttrT = TypeVar("_OpAttrT", bound=Attribute)
+
+OpAttr: TypeAlias = Annotated[_OpAttrT, IRDLAnnotations.AttributeDefAnnot]
+OptOpAttr: TypeAlias = Annotated[_OpAttrT | None,
+                                 IRDLAnnotations.OptAttributeDefAnnot]
+
+
 @dataclass(kw_only=True)
 class OpDef:
     """The internal IRDL definition of an operation."""
@@ -466,16 +500,107 @@ class OpDef:
                 "adding a 'name' field.")
 
         op_def = OpDef(clsdict["name"])
+        for field_name, field_type in get_type_hints(
+                pyrdl_def, include_extras=True).items():
 
-        for field_name, field_value in clsdict.items():
-            if isinstance(field_value, OperandDef):
-                op_def.operands.append((field_name, field_value))
-            elif isinstance(field_value, ResultDef):
-                op_def.results.append((field_name, field_value))
-            elif isinstance(field_value, RegionDef):
-                op_def.regions.append((field_name, field_value))
-            elif isinstance(field_value, AttributeDef):
-                op_def.attributes[field_name] = field_value
+            if field_name in get_type_hints(Operation).keys():
+                continue
+
+            # If the field type is an Annotated, separate the origin
+            # from the arguments.
+            # If the field type is not an Annotated, then the arguments should
+            # just be the field itself.
+            origin = get_origin(field_type)
+            args: tuple[Any, ...]
+            if origin is None:
+                args = (field_type, )
+            elif origin == Annotated:
+                args = get_args(field_type)
+            else:
+                args = (field_type, )
+            args = cast(tuple[Any, ...], args)
+
+            # Get attribute constraints from a list of pyrdl constraints
+            def get_constraint(
+                    pyrdl_constrs: tuple[Any, ...]) -> AttrConstraint:
+                constraints = [
+                    irdl_to_attr_constraint(pyrdl_constr)
+                    for pyrdl_constr in pyrdl_constrs
+                    if not isinstance(pyrdl_constr, IRDLAnnotations)
+                ]
+                if len(constraints) == 0:
+                    return AnyAttr()
+                if len(constraints) == 1:
+                    return constraints[0]
+                return AllOf(constraints)
+
+            # Get the operand, result, attribute, or region definition, from
+            # the pyrdl description.
+
+            # For operands and results, constrants are encoded as arguments of
+            # an Annotated, where the origin is the definition type (operand,
+            # optional result, etc...).
+            # For Attributes, constraints are encoded in the origin and the
+            # args of the Annotated, and the definition type (required or
+            # optional) is given in the Annotated arguments.
+            # For Regions, SingleBlock regions are given as Annotated arguments,
+            # and otherwise the Annotated origin (if it is an Annotated) gives
+            # the Region definition (required, optional, or variadic).
+
+            # Operand annotation
+            if args[0] == Operand:
+                constraint = get_constraint(args[1:])
+                op_def.operands.append((field_name, OperandDef(constraint)))
+            elif args[0] == list[Operand]:
+                constraint = get_constraint(args[1:])
+                op_def.operands.append((field_name, VarOperandDef(constraint)))
+            elif args[0] == (Operand | None):
+                constraint = get_constraint(args[1:])
+                op_def.operands.append((field_name, OptOperandDef(constraint)))
+
+            # Result annotation
+            elif args[0] == OpResult:
+                constraint = get_constraint(args[1:])
+                op_def.results.append((field_name, ResultDef(constraint)))
+            elif args[0] == list[OpResult]:
+                constraint = get_constraint(args[1:])
+                op_def.results.append((field_name, VarResultDef(constraint)))
+            elif args[0] == (OpResult | None):
+                constraint = get_constraint(args[1:])
+                op_def.results.append((field_name, OptResultDef(constraint)))
+
+            # Attribute annotation
+            elif IRDLAnnotations.AttributeDefAnnot in args:
+                constraint = get_constraint(args)
+                op_def.attributes[field_name] = AttributeDef(constraint)
+            elif IRDLAnnotations.OptAttributeDefAnnot in args:
+                assert get_origin(args[0]) in [UnionType, Union]
+                args = (reduce(lambda x, y: x | y,
+                               get_args(args[0])[:-1]), *args[1:])
+                constraint = get_constraint(args)
+                op_def.attributes[field_name] = OptAttributeDef(constraint)
+
+            # Region annotation
+            elif args[0] == Region:
+                if (len(args) > 1
+                        and args[1] == IRDLAnnotations.SingleBlockRegionAnnot):
+                    op_def.regions.append((field_name, SingleBlockRegionDef()))
+                else:
+                    op_def.regions.append((field_name, RegionDef()))
+            elif args[0] == VarRegion:
+                if (len(args) > 1
+                        and args[1] == IRDLAnnotations.SingleBlockRegionAnnot):
+                    op_def.regions.append(
+                        (field_name, VarSingleBlockRegionDef()))
+                else:
+                    op_def.regions.append((field_name, VarRegionDef()))
+            elif args[0] == OptRegion:
+                if (len(args) > 1
+                        and args[1] == IRDLAnnotations.SingleBlockRegionAnnot):
+                    op_def.regions.append(
+                        (field_name, OptSingleBlockRegionDef()))
+                else:
+                    op_def.regions.append((field_name, OptRegionDef()))
 
         op_def.options = clsdict.get("irdl_options", [])
 
@@ -1153,7 +1278,9 @@ def irdl_attr_try_builder(
     num_non_defaults = defaults.count(inspect.Signature.empty)
     if num_non_defaults > len(args):
         return None
-    for arg, param in zip(args, params[:num_non_defaults]):
+    if len(params) < len(args):
+        return None
+    for arg, param in zip(args, params[:len(args)]):
         if not is_satisfying_hint(arg, param):
             return None
     return builder(*args, *defaults[len(args):])
@@ -1169,8 +1296,7 @@ def irdl_attr_builder(cls: type[_PAttrT],
         res = irdl_attr_try_builder(builder, *args)
         if res:
             return res
-    raise TypeError(
-        f"No available {cls.__name__} builders for arguments {args}")
+    raise BuilderNotFoundException(cls, args)
 
 
 def irdl_param_attr_definition(cls: type[_PAttrT]) -> type[_PAttrT]:
