@@ -1,11 +1,12 @@
 from __future__ import annotations
 
+import json
 from dataclasses import dataclass, field
 from enum import Enum
 from frozenlist import FrozenList
 from typing import Iterable, TypeVar, Any, Dict, Optional, List, cast
 
-from xdsl.dialects.memref import MemRefType
+from xdsl.dialects.memref import AnyUnrankedMemrefType, MemRefType, UnrankedMemrefType
 from xdsl.ir import (BlockArgument, MLIRType, SSAValue, Block, Callable,
                      Attribute, Region, Operation, Data, ParametrizedAttribute)
 from xdsl.utils.diagnostic import Diagnostic
@@ -14,7 +15,8 @@ from xdsl.dialects.builtin import (
     DenseIntOrFPElementsAttr, Float16Type, Float32Type, Float64Type, FloatAttr,
     IndexType, IntegerType, NoneAttr, OpaqueAttr, Signedness, StringAttr,
     FlatSymbolRefAttr, IntegerAttr, ArrayAttr, IntAttr, TensorType, UnitAttr,
-    FunctionType, UnrankedTensorType, UnregisteredOp, VectorType)
+    FunctionType, UnrankedTensorType, UnregisteredOp, VectorType,
+    DictionaryAttr)
 
 indentNumSpaces = 2
 
@@ -77,12 +79,11 @@ class Printer:
     def _add_message_on_next_line(self, message: str, begin_pos: int,
                                   end_pos: int):
         """Add a message that will be displayed on the next line."""
-        # Python does not have a way to specify the type of an optional
-        # argument.
-        callback: Callable[[int], None] = (
-            lambda indent=self._indent: self._print_message(
-                message, begin_pos, end_pos, indent))
-        self._next_line_callback.append(cast(Callable[[], None], callback))
+
+        def callback(indent: int = self._indent):
+            self._print_message(message, begin_pos, end_pos, indent)
+
+        self._next_line_callback.append(callback)
 
     def _print_message(self,
                        message: str,
@@ -112,6 +113,8 @@ class Printer:
         self._print_new_line(indent=0, print_message=False)
 
     T = TypeVar('T')
+    K = TypeVar('K')
+    V = TypeVar('V')
 
     def print_list(self,
                    elems: Iterable[T],
@@ -122,6 +125,18 @@ class Printer:
                 self.print(delimiter)
             print_fn(elem)
 
+    def print_dictionary(self,
+                         elems: dict[K, V],
+                         print_key: Callable[[K], None],
+                         print_value: Callable[[V], None],
+                         delimiter: str = ", ") -> None:
+        for i, (key, value) in enumerate(elems.items()):
+            if i:
+                self.print(delimiter)
+            print_key(key)
+            self.print("=")
+            print_value(value)
+
     def _print_new_line(self,
                         indent: int | None = None,
                         print_message: bool = True) -> None:
@@ -281,6 +296,9 @@ class Printer:
         self.print_list(params, self.print_attribute)
         self.print(">")
 
+    def print_string_literal(self, string: str):
+        self.print(json.dumps(string))
+
     def print_attribute(self, attribute: Attribute) -> None:
         if isinstance(attribute, UnitAttr):
             return
@@ -309,7 +327,7 @@ class Printer:
                 return
 
         if isinstance(attribute, StringAttr):
-            self.print(f'"{attribute.data}"')
+            self.print_string_literal(attribute.data)
             return
 
         if isinstance(attribute, FlatSymbolRefAttr):
@@ -335,7 +353,8 @@ class Printer:
 
         if isinstance(attribute, FloatAttr):
             value = attribute.value
-            typ = attribute.type
+            typ = cast(FloatAttr[Float16Type | Float32Type | Float64Type],
+                       attribute).type
             self.print(value.data)
             self.print(" : ")
             self.print_attribute(typ)
@@ -349,6 +368,13 @@ class Printer:
             self.print_string("]")
             return
 
+        if isinstance(attribute, DictionaryAttr):
+            self.print_string("{")
+            self.print_dictionary(attribute.data, self.print_string_literal,
+                                  self.print_attribute)
+            self.print_string("}")
+            return
+
         # Function types have an alias in MLIR, but not in xDSL
         if (isinstance(attribute, FunctionType)
                 and self.target == self.Target.MLIR):
@@ -408,8 +434,9 @@ class Printer:
             attribute = cast(AnyVectorType, attribute)
             self.print(
                 "vector<" if isinstance(attribute, VectorType) else "tensor<")
-            self.print_list(attribute.shape.data,
-                            lambda x: self.print(x.value.data), "x")
+            self.print_list(
+                attribute.shape.data, lambda x: self.print(x.value.data)
+                if x.value.data != -1 else self.print("?"), "x")
             if len(attribute.shape.data) != 0:
                 self.print("x")
             self.print(attribute.element_type)
@@ -430,13 +457,23 @@ class Printer:
                 and self.target == self.Target.MLIR):
             attribute = cast(MemRefType[Attribute], attribute)
             self.print("memref<")
-            self.print_list(attribute.shape.data,
-                            lambda x: self.print(x.value.data), "x")
+            self.print_list(
+                attribute.shape.data, lambda x: self.print(x.value.data)
+                if x.value.data != -1 else self.print("?"), "x")
             self.print("x", attribute.element_type)
             self.print(">")
             return
 
-        # index type have an alias in MLIR, but not in xDSL
+        # Unranked tensors have an alias in MLIR, but not in xDSL
+        if (isinstance(attribute, UnrankedMemrefType)
+                and self.target == self.Target.MLIR):
+            attribute = cast(AnyUnrankedMemrefType, attribute)
+            self.print("memref<*x")
+            self.print(attribute.element_type)
+            self.print(">")
+            return
+
+        # IndexType has an alias in MLIR, but not in xDSL
         if (isinstance(attribute, IndexType)
                 and self.target == self.Target.MLIR):
             self.print("index")
@@ -509,7 +546,7 @@ class Printer:
         self.print(" ")
         self.print("[" if self.target == Printer.Target.XDSL else "{")
 
-        attribute_list = [p for p in attributes.items()]
+        attribute_list = list(attributes.items())
         self.print_list(attribute_list, self._print_attr_string)
 
         self.print("]" if self.target == Printer.Target.XDSL else "}")
